#ifndef PGAC_AUDIT_H
#define PGAC_AUDIT_H

#include "postgres.h"
#include "utils/acl.h"
#include <tcop/tcopprot.h>

/**
 * @file
 * @brief Спецификация подсистемы аудита
 * @ingroup iutils
 *
 * Основные принципы аудита, которые положены в основу реализации.
 *
 * В соответствии с РД подсистемой аудита должны регистрироваться
 * следующие виды событий:
 * -# использование идентификационного и аутентификационнго механизмов;
 * -# запросы на доступ к защищаемым ресурсам;
 * -# создание и уничтожение объектов;
 * -# действия по изменения ПРД.
 *
 * Практически все команды СУБД в процессе выполнения могут порождать
 * множество запросов на доступ, часть из которых может быть удовлетворено,
 * а часть --- нет. Более того, поскольку в СУБД обработка данных
 * производится в рамках транзакций, то даже успешное выполнение какой-либо
 * команды на модификацию системного каталога или данных не означает,
 * что выполненые ей изменения будут впоследствии зафиксированы в базе.
 *
 * Таким образом записи аудита можно рассматривать только как регистрацию
 * фактов появления запросов на доступ к объектам, но не фактов собственно
 * доступа.
 *
 * Соответственно оказывается невозможным зарегистрировать факты создания
 * и уничтожения объектов (так как в момент выполнения команды создания или
 * удаления нет возможности реально подтвердить факт создания или удаления,
 * а в момент подтверждения изменений отсутствует информация о выполненых
 * командах).
 *
 * В качестве решения этой проблемы реализована регистрация фактов запросов
 * на разрешение создания объектов в базе данных (пространстве имен).
 *
 * В лог аудита не выводится информации о возникновении ошибок выполнения,
 * не относяшихся к перечисленным событиям.
 *
 * Специальносу аудиту подвергаются команды создания и удаления пользователей
 * и групп пользователей.
 *
 * @todo Привести весь аудит к схеме регистрации успеха или неудачи "запроса"
 *       на доступ, а не самого доступа (рассматривать ли как запрос на доступ
 *       запрос на владение объектом и как их интерпретировать).
 *
 * @todo Выполнять проверку прав на модификацию схемы данных при удалении
 *       объектов?
 *
 * @todo Выполнять аудит проверок владения для всех объектов.
 *
 * @todo Аудит создания, удаления и модификации пользователей и групп.
 */


/**
 * Объявление флагов применения конфигураций
 */
extern bool pgac_applied_ignore_socket_maclabel;

/**
 * @name Конфигурационные переменные модуля
 */
/** @{*/

//!  Путь и имя конфигурационного файла аудита
/**
 *
 */
#define AUDIT_FILENAME	"pg_audit.conf"

extern char *AuditFileName;

/** @}*/

//! Статус сообщения
typedef enum pgac_Status {
	//! Успешное выполнение
	pgac_STATUS_SUCCESS = 0,
	//! Отказ в доступе
	pgac_STATUS_FAILURE = 1
} pgac_Status;

//! Аудит запросов к КСЗИ
/**
 * Одно действие может приводить к выводу в журнал аудита сразу
 * множества записей аудита.
 *
 * События показывают успешность или неуспешность запросов на
 * выполнение соответствующих действий с СУБД. При этом отслеживаются
 * только факты разрешения соотвествующих действий механизмом КСЗИ.
 * Прочие возможные ошибки не регистрируются (т.е. успешность выполнения
 * запроса на выполнение команды вовсе не означает ее успешное выполнение).
 *
 * Все флаги разбиты на смысловые группы, каждой из которых сопоставлена
 * соответствующая шестнадцатиричная цифра (или группа цифр).
 */
typedef enum pgac_Event {
	/**
	 * @name Регистрация запросов на разрешение выполнения команд СУБД и проверки прав доступа
	 *
	 * Эта группа флагов использует самые младшие 4-е бита (первую справа
	 * шестнадцатиричную цифру).
	 */
	/** @{*/
	//! Запрос на модификацию субъектов кластера
	pgac_EVENT_SUBJECT       = (1 << 0),
	//! Запрос на модификацию конфигурации сессии, относящейся к КСЗИ
	pgac_EVENT_CONFIGURATION = (1 << 1),
	//! Запрос на модификацию прав доступа к объектам БД
	pgac_EVENT_RIGHTS        = (1 << 2),
	//! Deprecated
	pgac_EVENT_CHECK_RIGHTS	 = (1 << 3),
	/** @}*/

	/**
	 * @name Регистрация выполнения запросов к данным
	 *
	 * Эта группа использует следующие 4-ре бита
	 */
	/** @{*/
	//! SELECT
	pgac_EVENT_SELECT      = (1 << 4),
	//! INSERT
	pgac_EVENT_INSERT      = (1 << 5),
	//! UPDATE
	pgac_EVENT_UPDATE      = (1 << 6),
	//! DELETE
	pgac_EVENT_DELETE      = (1 << 7),
	/** @}*/

	/**
	 * @name Регистрация специальных операций с данными и создания правил контроля ссылочной целостности
	 *
	 * Эта группа использует следующие 4-ре бита
	 */
	/** @{*/
	//! TRUNCATE
	pgac_EVENT_TRUNCATE    = (1 << 8),

	//! REFERENCES
	pgac_EVENT_REFERENCES  = (1 << 10),
	//! TRIGGER
	pgac_EVENT_TRIGGER     = (1 << 11),
	/** @}*/

	/**
	 * @name Регистарция использования/запуска объектов
	 *
	 * Занимает следующие 4-ре бита
	 */
	/** @{*/
	//! EXECUTE
	pgac_EVENT_EXECUTE     = (1 << 12),
	//! USAGE
	pgac_EVENT_USAGE       = (1 << 13),
	/** @}*/

	/**
	 * @name Регистрация создания/удаления и модификации объектов
	 *
	 * Занимает следующие 4-ре бита
	 */
	/** @{*/
	//! CREATE
	pgac_EVENT_CREATE      = (1 << 16),
	//! CREATE TEMP
	pgac_EVENT_CREATE_TEMP = (1 << 17),
	//! DROP
	/**
	 * Это право неявно принадлежит
	 * владельцам объектов и суперпользователям и не может быть передано.
	 */
	pgac_EVENT_DROP = (1 << 18),
	//! ALTER
	/**
	 * Аналогично "праву" DROP это право неявно принадлежит владельцам объектов
	 * и суперпользователям и не может быть передано.
	 */
	pgac_EVENT_ALTER = (1 << 19),
	/** @}*/

	/**
	 * @name Регистрация использования механизма идентификации/аутентификации
	 *
	 * Для этой группы флагов используются старшие 4 бита (последнее справа
	 * шестнадцатирично число). Эти флаги включены всегда, вне зависимости
	 * от настроек аудита, поэтому при конфигурировании аудита их можно не
	 * учитывать.
	 */
	/** @{*/
	//! Соединение с базой данных
	pgac_EVENT_CONNECT    = (1 << 30),
	//! Расоединение с базой данных
	pgac_EVENT_DISCONNECT = (1 << 31)
	/** @}*/
} pgac_Event;

//! Вывод информации в лог
/**
 * @param _s статус сообщения
 * @param _e событие, вызвавшее это сообщение
 * @param _u идентификатор субъекта доступа
 * @param _f формат вывода дополнительной информации о доступе
 */
void pgac_log (pgac_Status _s, pgac_Event _e, Oid _u, const char *_f, ...)
		  __attribute__((format(PG_PRINTF_ATTRIBUTE, 4, 5)));

//! Вывод информации в журнал регистрации событий
/**
 * @param status статус сообщения
 * @param event событие, вызвавшее это сообщение
 * @param roleId идентификатор субъекта доступа
 * @param print_sql признак вывода SQL
 * @param fmt формат вывода дополнительной информации о доступе
 */
#define pgac_log(status, event, roleId, fmt, args...) \
	pgac_log_sql(status, event, roleId, false, fmt, ## args)

#define pgac_sqllog(status, event, roleId, fmt, args...) \
	pgac_log_sql(status, event, roleId, true, fmt, ## args)

void pgac_log_sql (pgac_Status status, pgac_Event event, Oid roleId, bool print_sql, const char *fmt, ...)
	__attribute__((format(PG_PRINTF_ATTRIBUTE, 5, 6)));

/* Логирование успешного подключения */
void pgac_log_successful_connection(void);

#define pgac_debug_aclcheck(status, objkind, mask, roleId, objname, colname) {if (pgac_debug_print) \
{ pgac_aclcheck_log (status, objkind, mask, roleId, objname, colname); };}

void pgac_aclcheck_log (pgac_Status status, AclObjectKind objkind, AclMode mask, Oid roleId,
						  const char *objname, const char *colname);

const char *pgac_printAclObjectKind(AclObjectKind objkind);

const char *pgac_printObjectType(ObjectType objtype);

const char *pgac_printTriggerType(int tgtype_event);

pgac_Event pgac_AclMask2Event(AclMode mask, bool CheckRead);

//! Инициализация сессии аудита
void pgac_audit_init (void);

//! Завершение работы модуля аудита
void pgac_audit_done (void);

#endif /* PGAC_AUDIT_H */
